#!/usr/bin/env python3
"""Test model training workflows."""

import json
from pathlib import Path

import numpy as np
import pytest
from ctapipe.io import TableLoader, read_table
from ctapipe.reco.reconstructor import ReconstructionProperty, Reconstructor

from datapipe.tests.utils import run_cwl

TRAINING_CONFIG = """
TrainEnergyRegressor:
  CrossValidator:
    n_cross_validations: 5

  EnergyRegressor:
    model_cls: ExtraTreesRegressor
    log_target: True

    model_config:
      n_estimators: 10
      max_depth: 10
      n_jobs: -1

    features:
      - hillas_intensity
      - HillasReconstructor_tel_impact_distance

TrainParticleClassifier:
  CrossValidator:
    n_cross_validations: 5

  ParticleClassifier:
    model_cls: ExtraTreesClassifier
    model_config:
      n_estimators: 10
      max_depth: 10
      n_jobs: -1

    features:
      - hillas_intensity
      - hillas_width
      - HillasReconstructor_tel_impact_distance

TrainDispReconstructor:
  CrossValidator:
    n_cross_validations: 5

  DispReconstructor:
    norm_cls: ExtraTreesRegressor
    sign_cls: ExtraTreesClassifier
    norm_config:
      n_estimators: 10
      max_depth: 10
      n_jobs: -1
    sign_config:
      n_estimators: 10
      max_depth: 10
      n_jobs: -1

    features:
      - hillas_intensity
      - hillas_width
      - hillas_r
"""


@pytest.fixture
def training_config_file(tmp_path):
    """Create a YAML config file for use with ctapipe training tools."""
    training_config_file = tmp_path / "train_energy.conf.yaml"
    training_config_file.write_text(TRAINING_CONFIG)
    return training_config_file


@pytest.mark.verifies_usecase("DPPS-UC-130-1.3.1")
def test_train_energy_model(tmp_path, gammas_dl2_only_geom_path, training_config_file):
    """Test training an energy regressor."""
    gammas = gammas_dl2_only_geom_path
    model_path = tmp_path / "energy_reco_model.pkl"  # the default filename
    cross_validation_path = tmp_path / "energy_cross_validation.h5"

    job_file = tmp_path / "train_energy_model-job.json"
    job_file.write_text(
        json.dumps(
            dict(
                config={"class": "File", "path": str(training_config_file.name)},
                gamma_events={"class": "File", "path": str(gammas)},
            ),
            indent=2,
        )
    )

    result = run_cwl(
        Path("workflows/train_energy_model.cwl").absolute(),
        inputs_path=job_file,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    assert model_path.exists(), f"expected model file: {model_path}"
    assert cross_validation_path.exists()

    # Success item 1: try to load the model and reconstruct some events
    #
    with TableLoader(gammas) as loader:
        reco = Reconstructor.from_name(
            "EnergyRegressor",
            subarray=loader.subarray,
            model_cls="ExtraTreesRegressor",
            load_path=model_path,
        )
        data = loader.read_telescope_events_by_id(1)[1]
        result = reco.predict_table(loader.subarray.telescope_types[0], data)

        assert ReconstructionProperty.ENERGY in result

        prediction = result[ReconstructionProperty.ENERGY]

        # Now just check that there are values and that they have some
        # reasonable mean
        parameter = "ExtraTreesRegressor_tel_energy"
        assert parameter in prediction.colnames
        assert np.count_nonzero(prediction[parameter]) > 10
        assert np.nanmean(prediction[parameter]) > 0.1

    # Success item 2: check cross_validation exists:
    for tel in loader.subarray.telescope_types:
        cv_table = read_table(cross_validation_path, f"/cv_predictions/{tel}")
        assert len(cv_table) > 1000
        assert "ExtraTreesRegressor_energy" in cv_table.columns


@pytest.mark.verifies_usecase("DPPS-UC-130-1.3.2")
def test_train_gammaness_model(
    tmp_path,
    gammas_dl2_only_geom_path,
    protons_dl2_path,
    training_config_file,
):
    """Test training an energy regressor."""
    gammas = gammas_dl2_only_geom_path
    protons = protons_dl2_path
    cross_validation_path = tmp_path / "gammaness_cross_validation.h5"

    job_file = tmp_path / "train_gammaness_model-job.json"
    job_file.write_text(
        json.dumps(
            dict(
                config={"class": "File", "path": str(training_config_file.name)},
                gamma_events={"class": "File", "path": str(gammas)},
                background_events={"class": "File", "path": str(protons)},
            ),
            indent=2,
        )
    )

    model_path = tmp_path / "gammaness_reco_model.pkl"  # the default filename
    result = run_cwl(
        Path("workflows/train_gammaness_model.cwl").absolute(),
        inputs_path=job_file,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    # Success item 1: load the events and see if we can predict

    with TableLoader(gammas) as loader:
        reco = Reconstructor.from_name(
            "ParticleClassifier",
            subarray=loader.subarray,
            model_cls="ExtraTreesClassifier",
            load_path=model_path,
        )
        data = loader.read_telescope_events_by_id(1)[1]
        result = reco.predict_table(loader.subarray.telescope_types[0], data)

        assert ReconstructionProperty.PARTICLE_TYPE in result

        prediction = result[ReconstructionProperty.PARTICLE_TYPE]

        # Now just check that there are values and that they have some
        # reasonable mean
        parameter = "ExtraTreesClassifier_tel_prediction"
        assert parameter in prediction.colnames
        assert np.count_nonzero(prediction[parameter]) > 10
        assert np.nanmean(prediction[parameter]) > 0.1

    # Success item 2: check cross_validation exists:
    for tel in loader.subarray.telescope_types:
        cv_table = read_table(cross_validation_path, f"/cv_predictions/{tel}")
        assert len(cv_table) > 1000
        assert "ExtraTreesClassifier_prediction" in cv_table.columns


@pytest.mark.verifies_usecase("DPPS-UC-130-1.3.3")
def test_train_disp_model(tmp_path, gammas_dl2_only_geom_path, training_config_file):
    """Test training an energy regressor."""
    gammas = gammas_dl2_only_geom_path
    cross_validation_path = tmp_path / "disp_cross_validation.h5"

    job_file = tmp_path / "train_disp_model-job.json"
    job_file.write_text(
        json.dumps(
            dict(
                config={"class": "File", "path": str(training_config_file.name)},
                gamma_events={"class": "File", "path": str(gammas)},
            ),
            indent=2,
        )
    )

    model_path = tmp_path / "disp_reco_model.pkl"  # the default filename
    result = run_cwl(
        Path("workflows/train_disp_model.cwl").absolute(),
        inputs_path=job_file,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    # Success item 1: try to load it back up and predict

    with TableLoader(gammas) as loader:
        reco = Reconstructor.from_name(
            "DispReconstructor",
            subarray=loader.subarray,
            norm_cls="ExtraTreesRegressor",
            sign_cls="ExtraTreesClassifier",
            load_path=model_path,
        )
        data = loader.read_telescope_events_by_id(1)[1]
        result = reco.predict_table(loader.subarray.telescope_types[0], data)

        assert ReconstructionProperty.DISP in result

        prediction = result[ReconstructionProperty.DISP]

        # Now just check that there are values and that they have some
        # reasonable mean
        parameter = "disp_tel_parameter"
        assert parameter in prediction.colnames
        assert np.count_nonzero(prediction[parameter]) > 10

    # Success item 2: check cross_validation exists:

    for tel in loader.subarray.telescope_types:
        cv_table = read_table(cross_validation_path, f"/cv_predictions/{tel}")
        assert len(cv_table) > 1000
        assert "disp_parameter" in cv_table.columns
        assert "disp_sign_score" in cv_table.columns



#!/usr/bin/env python3

"""Configure tests for CWL."""

import pytest
from ctapipe.utils import get_dataset_path

URL = "https://minio-cta.zeuthen.desy.de/dpps-testdata-public/data/datapipe-test-data/"


@pytest.fixture(scope="session")
def gammas_dl2_path():
    """Sample DL2 gammas file."""
    return get_dataset_path("gamma_test_geo_en_cl.dl2.h5", url=URL)


@pytest.fixture(scope="session")
def gammas_dl2_only_geom_path():
    """Sample DL2 gammas file with only reconstruccted geometry columns.."""
    return get_dataset_path("gamma_test_geo.dl2.h5", url=URL)


@pytest.fixture(scope="session")
def electrons_dl2_path():
    """Sample DL2 electrons file."""
    return get_dataset_path("electron_test_geo_en_cl.dl2.h5", url=URL)


@pytest.fixture(scope="session")
def protons_dl2_path():
    """Sample DL2 protons file."""
    return get_dataset_path("proton_test_geo_en_cl.dl2.h5", url=URL)


@pytest.fixture(scope="session")
def event_selection_path():
    """Sample event cuts file as output from ctapipe-optimize-event-selection."""
    return get_dataset_path("event_selection.fits", url=URL)


@pytest.fixture(scope="session")
def energy_model_path():
    """Sample gamma energy reconstruction model."""
    return get_dataset_path("energy_model.pkl", url=URL)


@pytest.fixture(scope="session")
def classifier_model_path():
    """Sample gammaness reconstruction model."""
    return get_dataset_path("classifier_model.pkl", url=URL)



#!/usr/bin/env python3

"""Tests of IRF workflows."""

import json
from pathlib import Path

import numpy as np
import pytest
from ctapipe.io import TableLoader

from datapipe.tests.utils import run_cwl


@pytest.mark.verifies_usecase("DPPS-UC-130-1.4")
def test_apply(
    tmp_path,
    gammas_dl2_only_geom_path: Path,
    energy_model_path: Path,
    classifier_model_path: Path,
):
    """Test optimize UC."""
    config_file = tmp_path / "apply_models-job.json"
    out_file = "applied.dl2.h5"
    config_file.write_text(
        json.dumps(
            dict(
                events={"class": "File", "path": str(gammas_dl2_only_geom_path)},
                models=[
                    {"class": "File", "path": str(energy_model_path)},
                    {"class": "File", "path": str(classifier_model_path)},
                ],
                output_filename=out_file,
            ),
            indent=2,
        )
    )

    out_path = tmp_path / out_file

    result = run_cwl(
        Path("workflows/apply_models.cwl").absolute(),
        inputs_path=config_file,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    with TableLoader(out_path, dl2=True) as loader:
        sub_events = loader.read_subarray_events()
        assert len(sub_events) > 5, "No or too few events in output"

        for col in [
            "RandomForestRegressor_energy",
            "RandomForestClassifier_prediction",
        ]:
            assert col in sub_events.colnames
            assert len(np.isfinite(sub_events[col])) > 0, "values are all NaN"



"""Tests for the cwl workflow for processing multiple files and then merge them."""

import logging
import subprocess as sp
from pathlib import Path

import yaml
from ctapipe.utils import get_dataset_path

from datapipe.tests.utils import run_cwl

log = logging.getLogger(__name__)


def test_workflow_process_multiple(tmp_path):
    """Test cwl workflow for processing multiple files, then merge."""
    input_file = get_dataset_path("gamma_prod5.simtel.zst")
    # create to dl1 files with different obs_ids, merge checks for
    # same subarray and different obs-ids, but we do not have two test files
    # that are small and similar
    inputs = {
        "input_files": [],
        "output_filename": "merged.dl1.h5",
    }
    for obs_id in (1, 2):
        output_path = tmp_path / f"gamma_{obs_id}.dl1_img.h5"
        sp.run(
            [
                "ctapipe-process",
                f"--input={input_file}",
                f"--output={output_path}",
                f"--SimTelEventSource.override_obs_id={obs_id}",
                "--write-images",
                "--no-write-parameters",
            ],
            check=True,
        )
        inputs["input_files"].append({"class": "File", "path": str(output_path)})

    inputs_path = tmp_path / "inputs.yaml"
    with inputs_path.open("w") as f:
        yaml.dump(inputs, f)

    cwl_path = tmp_path / "cwl_run"
    cwl_path.mkdir()

    result = run_cwl(
        Path("workflows/process_dl0_dl1_multiple.cwl").absolute(),
        inputs_path=inputs_path,
        cwd=cwl_path,
    )
    log.info("cwltool output:\n%s", result.stdout)
    assert result.returncode == 0, result.stdout
    assert (cwl_path / "merged.dl1.h5").is_file()
    assert (cwl_path / "ctapipe-merge.provenance.log").is_file()

    for obs_id in (1, 2):
        name = f"gamma_{obs_id}.dl1.h5"
        # intermediate output files shouldn't be in the output
        assert not (cwl_path / name).is_file()
        # but log and provlog should
        assert (cwl_path / f"{name}.provlog").is_file()



"""DataPipe workflow tests for CLIs and simple workflows."""

import logging
import subprocess as sp
from pathlib import Path

import pytest
import yaml
from astropy.utils.data import shutil
from ctapipe.io import TableLoader
from ctapipe.utils import get_dataset_path

from datapipe.tests.utils import run_cwl

LOG = logging.getLogger("test_workflows")


@pytest.mark.verifies_usecase("DPPS-UC-130-1.2.1")
def test_dl0_dl1_main(tmp_path):
    """Test transformation from DL0 to DL1."""
    values = tmp_path / "values.cfg"
    values.write_text(
        yaml.dump(
            dict(
                dl0="dataset://gamma_prod5.simtel.zst",
                dl1_filename="events.dl1.h5",
            )
        )
    )

    result = run_cwl(
        Path("workflows/process_dl0_dl1.cwl").absolute(),
        inputs_path=values,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    # Now check that the events.dl1.h5 file looks ok.

    out_path = tmp_path / "events.dl1.h5"
    assert out_path.exists()

    with TableLoader(out_path) as loader:
        assert len(loader.read_scheduling_blocks()) > 0

        sub_events = loader.read_subarray_events()
        assert len(sub_events) > 5
        assert "time" in sub_events.colnames

        tel_events = loader.read_telescope_events()
        assert len(tel_events) > 5
        assert "tel_id" in tel_events.colnames
        assert "hillas_width" in tel_events.colnames


@pytest.mark.verifies_usecase("DPPS-UC-130-1.2.1")
def test_dl0_dl1_camcalib(tmp_path):
    """Test transformation from DL0 to DL1 with a camcalib monitoring file."""
    dl0_path = get_dataset_path("gamma_prod6_preliminary.simtel.zst")
    camera_calibration_path = get_dataset_path(
        "calibpipe_camcalib_single_chunk_i0.1.0.dl1.h5"
    )

    dl1_filename = "events_with_camcalib.dl1.h5"
    values = tmp_path / "values.cfg"
    config = tmp_path / "config.yaml"
    allowed_tels = [1]

    with config.open("w") as f:
        yaml.dump(
            {
                "SimTelEventSource": {
                    "allowed_tels": allowed_tels,
                    "skip_r1_calibration": True,
                },
            },
            f,
        )

    # use relative path to check if input is properly accessed
    shutil.copy2(dl0_path, tmp_path / dl0_path.name)
    shutil.copy2(camera_calibration_path, tmp_path / camera_calibration_path.name)

    values = tmp_path / "values.cfg"
    values.write_text(
        yaml.dump(
            {
                "dl0": {"class": "File", "path": dl0_path.name},
                "dl1_filename": dl1_filename,
                "camera_calibration_file": {
                    "class": "File",
                    "path": camera_calibration_path.name,
                },
                "processing_config": {"class": "File", "path": str(config)},
                # parameters doesn't work currently, as skip_r1_calibration implies no gain selection
                "write_parameters": False,
            }
        )
    )

    result = run_cwl(
        Path("workflows/process_dl0_dl1.cwl").absolute(),
        inputs_path=values,
        cwd=tmp_path,
    )
    LOG.info("Output of running cwl:\n%s", result.stdout)

    assert result.returncode == 0, result.stdout

    # Now check that the events_with_camcalib.dl1.h5 file looks ok.

    out_path = tmp_path / "events_with_camcalib.dl1.h5"
    assert out_path.exists()

    with TableLoader(out_path) as loader:
        assert len(loader.read_scheduling_blocks()) > 0

        tel_events = loader.read_telescope_events(dl1_images=True)
        assert len(tel_events) == 3
        assert "tel_id" in tel_events.colnames
        assert "image" in tel_events.colnames
        # check that we correctly applied calibrations.
        assert tel_events["image"].mean() > 0


@pytest.mark.verifies_usecase("DPPS-UC-130-1.2.2")
def test_dl1_dl2_main(tmp_path):
    """Test transformation from DL1 to DL2 up to stereo geometry (main scenario).

    This actually goes from DL0 to DL2, to avoid having to do a separate
    workflows for each step. However, the full workflow is tested in the
    following test.
    """
    values = tmp_path / "values.cfg"
    values.write_text(
        yaml.dump(
            dict(
                dl1="dataset://gamma_prod5.simtel.zst",
                dl2_filename="events.dl2.h5",
            )
        )
    )

    result = run_cwl(
        Path("workflows/process_dl1_dl2.cwl").absolute(),
        inputs_path=values,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    # Now check that the events.dl1.h5 file looks ok.

    out_path = tmp_path / "events.dl2.h5"
    assert out_path.exists()

    with TableLoader(out_path) as loader:
        assert len(loader.read_scheduling_blocks()) > 0

        sub_events = loader.read_subarray_events()
        assert len(sub_events) > 5
        assert "time" in sub_events.colnames
        assert "HillasReconstructor_h_max" in sub_events.colnames


@pytest.mark.xfail
@pytest.mark.verifies_usecase("DPPS-UC-130-1.2.2")
def test_dl1_dl2_extended(tmp_path):
    """Test transformation from DL1 to DL2 with extended reconstruction."""
    pytest.fail("to be implemented")


@pytest.mark.xfail
@pytest.mark.verifies_usecase("DPPS-UC-130-1.2.2")
def test_dl1_dl2_alt_mono(tmp_path):
    """Test transformation from DL1 to DL2 alternate scenario: mono reco."""
    pytest.fail("to be implemented")


@pytest.mark.verifies_usecase("DPPS-UC-130-1.2")
def test_workflow_dl0_dl2_main(tmp_path):
    """Test workflow of 2 steps: DL0 to DL1 followed by DL1 to DL2.

    Note this is only a PARTIAL validation of UC-130-1.2. The rest should be by
    inspection, since it requires a code review.
    """
    dl0_filename = str(get_dataset_path("gamma_prod5.simtel.zst"))

    values = tmp_path / "values.cfg"
    values.write_text(
        yaml.dump(
            {
                "dl0": {"class": "File", "path": dl0_filename},
            }
        )
    )

    result = run_cwl(
        Path("workflows/workflow_dl0_to_dl2.cwl").absolute(),
        inputs_path=values,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    # Now check that the events.dl1.h5 file looks ok.

    assert (tmp_path / "gamma_prod5.dl1.h5").exists()
    assert (tmp_path / "gamma_prod5.dl2.h5").exists()


@pytest.mark.verifies_usecase("DPPS-UC-130-1.8")
def test_workflow_merge(tmp_path):
    """Test cwl of merge tool."""
    input_file = get_dataset_path("gamma_prod5.simtel.zst")
    # create to dl1 files with different obs_ids, merge checks for
    # same subarray and different obs-ids, but we do not have two test files
    # that are small and similar
    inputs = {"input_files": [], "output_filename": "merged.dl1.h5"}
    for obs_id in (1, 2):
        output_path = tmp_path / f"gamma_{obs_id}.dl1.h5"
        sp.run(
            [
                "ctapipe-process",
                f"--input={input_file}",
                f"--output={output_path}",
                f"--SimTelEventSource.override_obs_id={obs_id}",
            ],
            cwd=tmp_path,
            check=True,
        )
        inputs["input_files"].append({"class": "File", "path": str(output_path)})

    inputs_path = tmp_path / "inputs.yaml"
    with inputs_path.open("w") as f:
        yaml.dump(inputs, f)

    result = run_cwl(
        Path("workflows/merge.cwl").absolute(),
        inputs_path=inputs_path,
        cwd=tmp_path,
    )
    assert result.returncode == 0, result.stdout
    assert (tmp_path / "merged.dl1.h5").is_file()
    assert (tmp_path / "ctapipe-merge.provenance.log").is_file()


@pytest.mark.verifies_usecase("DPPS-UC-130-1.3")
def test_dummy():
    """Hack to make UCs that are marked as by inspection pass."""
    assert True



#!/usr/bin/env python3
"""Sphinx directives to document a CWL workflows."""

import hashlib
import os
import re
import subprocess
from pathlib import Path

import yaml
from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.statemachine import ViewList
from sphinx.util.logging import getLogger
from sphinx.util.nodes import nested_parse_with_titles

logger = getLogger(__name__)

GRAPH_DIR = "_static"


def parse_rst(text, directive):
    """Parse RST string into nodes using Sphinx's parser."""
    result = ViewList()
    for line in text.splitlines():
        result.append(line, "<cwl-doc>")
    node = nodes.container()
    nested_parse_with_titles(directive.state, result, node)
    return node.children


class CWLWorkflowGraphDirective(Directive):
    """CWL Workflow Graph Directive."""

    has_content = False
    required_arguments = 1  # path to the CWL file

    def run(self):
        """Generate sphinx."""
        env = self.state.document.settings.env
        app = env.app

        cwl_rel_path = self.arguments[0]
        cwl_abs_path = os.path.abspath(os.path.join(env.srcdir, cwl_rel_path))

        if not os.path.isfile(cwl_abs_path):
            error = self.state_machine.reporter.error(
                f"CWL file not found: {cwl_rel_path}", line=self.lineno
            )
            return [error]

        # Unique output name based on file path
        hash_id = hashlib.sha256(cwl_rel_path.encode("utf-8")).hexdigest()[:10]
        base_name = os.path.splitext(os.path.basename(cwl_rel_path))[0]
        svg_filename = f"{base_name}-{hash_id}.svg"

        # Output directory for graphs
        graph_dir = os.path.join(app.srcdir, GRAPH_DIR, "cwl-graphs")
        os.makedirs(graph_dir, exist_ok=True)
        svg_path = os.path.join(graph_dir, svg_filename)

        # Run cwltool | dot -Tsvg
        try:
            logger.info(
                "[CWLGraph] Generating SVG for } %s -> %s", cwl_rel_path, svg_path
            )
            cwl_proc = subprocess.Popen(
                ["cwltool", "--print-dot", cwl_abs_path],
                stdout=subprocess.PIPE,
            )
            subprocess.run(
                ["dot", "-Tsvg", "-o", svg_path],
                stdin=cwl_proc.stdout,
                check=True,
            )
            cwl_proc.stdout.close()
            cwl_proc.wait()
        except subprocess.CalledProcessError as e:
            error = self.state_machine.reporter.error(
                f"Error generating workflow graph: {e}", line=self.lineno
            )
            return [error]

        # Embed the resulting SVG
        rel_svg_uri = os.path.join(GRAPH_DIR, "cwl-graphs", svg_filename)
        return [nodes.image(uri=rel_svg_uri)]


def slugify(text):
    """Normalize a name to work as a sphinx ID."""
    return re.sub(r"[^a-z0-9]+", "-", text.lower()).strip("-")


class CWLWorkflowDirective(Directive):
    """CWL Workflow Directive."""

    has_content = False
    required_arguments = 1  # path to the CWL file

    def run(self):
        """Generate sphinx."""
        env = self.state.document.settings.env
        cwl_rel_path = Path(self.arguments[0])
        cwl_abs_path = os.path.abspath(os.path.join(env.srcdir, cwl_rel_path))
        cwl_workflow_name = cwl_rel_path.name

        if not os.path.isfile(cwl_abs_path):
            error = self.state_machine.reporter.error(
                f"CWL file not found: {cwl_rel_path}", line=self.lineno
            )
            return [error]

        try:
            # Parse the whole CWL content here
            with open(cwl_abs_path) as f:
                content = yaml.safe_load(f)

            label = content.get("label", cwl_workflow_name.replace(".cwl", ""))
            doc = content.get("doc", None)
            inputs = content.get("inputs", {})
            outputs = content.get("outputs", {})
            cwl_class = content.get("class", "?")

            if isinstance(inputs, list):
                inputs = {item["id"]: item for item in inputs}
            if isinstance(outputs, list):
                outputs = {item["id"]: item for item in outputs}

        except Exception as e:
            error = self.state_machine.reporter.error(
                f"Failed to parse CWL file: {e}", line=self.lineno
            )
            return [error]

        # Create the document node
        section = nodes.section()
        section["ids"].append(slugify(cwl_workflow_name))

        section += nodes.title(text=label)
        info = nodes.paragraph()
        info += nodes.strong(text=cwl_class + ": ")
        info += nodes.Text(cwl_workflow_name)

        section += info
        section += parse_rst(doc, self)

        section += nodes.strong(text="Inputs:")
        section += self.make_table(inputs)

        section += nodes.strong(text="Outputs:")
        section += self.make_table(outputs)

        return [section]

    def make_table(self, io_dict):
        """Make table of inputs/outputs."""
        table = nodes.table()
        tgroup = nodes.tgroup(cols=3)
        table += tgroup

        for _ in range(3):
            tgroup += nodes.colspec(colwidth=1)

        thead = nodes.thead()
        tgroup += thead
        header_row = nodes.row()
        for label in ["ID", "Type", "Description"]:
            entry = nodes.entry()
            entry += nodes.paragraph(text=label)
            header_row += entry
        thead += header_row

        tbody = nodes.tbody()
        tgroup += tbody

        for key, val in io_dict.items():
            row = nodes.row()

            # ID
            id_entry = nodes.entry()
            id_entry += nodes.literal(text=key)
            row += id_entry

            # Type
            type_text = val.get("type", "Unknown")
            if isinstance(type_text, list):
                type_text = " | ".join(str(t) for t in type_text)
            type_entry = nodes.entry()
            type_entry += nodes.literal(text=str(type_text))
            row += type_entry

            # Description
            desc = val.get("doc", "")
            desc_entry = nodes.entry()
            desc_entry += parse_rst(desc, self)
            row += desc_entry

            tbody += row

        return table


def setup(app):
    """Create Sphinx App."""
    app.add_directive("cwl_workflow_graph", CWLWorkflowGraphDirective)
    app.add_directive("cwl_workflow", CWLWorkflowDirective)
    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Utils for datapipe testing"""

import subprocess as sp


def run_cwl(workflow_path, inputs_path=None, cwd=None):
    """Run cwltool on a workflow using subprocess."""
    command = ["cwltool", "--debug", str(workflow_path)]
    if inputs_path is not None:
        command.append(str(inputs_path))
    return sp.run(command, stdout=sp.PIPE, stderr=sp.STDOUT, text=True, cwd=cwd)



#!/usr/bin/env python3

"""Tests of IRF workflows."""

import json
from pathlib import Path

import pytest
from astropy.table import Table

from datapipe.tests.utils import run_cwl


@pytest.mark.verifies_usecase("DPPS-UC-130-1.9")
def test_optimize(
    tmp_path, gammas_dl2_path: Path, protons_dl2_path: Path, electrons_dl2_path: Path
):
    """Test optimize UC."""
    config_file = tmp_path / "optimize-job.json"
    config_file.write_text(
        json.dumps(
            dict(
                gammas={"class": "File", "path": str(gammas_dl2_path)},
                electrons={"class": "File", "path": str(electrons_dl2_path)},
                protons={"class": "File", "path": str(protons_dl2_path)},
                output_filename="event_selection.fits",
            ),
            indent=2,
        )
    )

    result = run_cwl(
        Path("workflows/optimize.cwl").absolute(),
        inputs_path=config_file,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    # look inside the output file for expected HDUs:
    for hdu in ["RAD_MAX", "GH_CUTS", "VALID_ENERGY", "VALID_OFFSET"]:
        Table.read(tmp_path / "event_selection.fits", hdu=hdu)


@pytest.mark.verifies_usecase("DPPS-UC-130-1.6")
def test_compute_irfs(
    tmp_path,
    gammas_dl2_path: Path,
    protons_dl2_path: Path,
    electrons_dl2_path: Path,
    event_selection_path: Path,
):
    """Test compute irfs UC."""
    config_file = tmp_path / "compute_irf-job.json"
    config_file.write_text(
        json.dumps(
            dict(
                gammas={"class": "File", "path": str(gammas_dl2_path)},
                electrons={"class": "File", "path": str(electrons_dl2_path)},
                protons={"class": "File", "path": str(protons_dl2_path)},
                event_selection={"class": "File", "path": str(event_selection_path)},
                output_filename="irf.fits.gz",
                output_metrics_filename="performance_metrics.fits",
            ),
            indent=2,
        )
    )

    result = run_cwl(
        Path("workflows/compute_irf.cwl").absolute(),
        inputs_path=config_file,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    # look inside the output file for expected HDUs:
    for hdu in ["EFFECTIVE AREA", "ENERGY DISPERSION", "PSF", "BACKGROUND"]:
        Table.read(tmp_path / "irf.fits.gz", hdu=hdu)

    for hdu in ["SENSITIVITY", "ANGULAR RESOLUTION"]:
        tab = Table.read(tmp_path / "performance_metrics.fits", hdu=hdu)
        assert "ENERG_LO" in tab.colnames


def test_both(
    tmp_path,
    gammas_dl2_path: Path,
    protons_dl2_path: Path,
    electrons_dl2_path: Path,
    event_selection_path: Path,
):
    """Test a workflow combining Optimize and Compute."""
    config_file = tmp_path / "workflow_optimize_and_irf-job.json"
    config_file.write_text(
        json.dumps(
            dict(
                gammas_optimize={"class": "File", "path": str(gammas_dl2_path)},
                electrons_optimize={"class": "File", "path": str(electrons_dl2_path)},
                protons_optimize={"class": "File", "path": str(protons_dl2_path)},
                gammas_irf={"class": "File", "path": str(gammas_dl2_path)},
                electrons_irf={"class": "File", "path": str(electrons_dl2_path)},
                protons_irf={"class": "File", "path": str(protons_dl2_path)},
                event_selection={"class": "File", "path": str(event_selection_path)},
                analysis_name="ctao-example",
            ),
            indent=2,
        )
    )

    result = run_cwl(
        Path("workflows/workflow_optimize_and_irf.cwl").absolute(),
        inputs_path=config_file,
        cwd=tmp_path,
    )

    assert result.returncode == 0, result.stdout

    assert (tmp_path / "ctao-example.irf.fits").exists()
    assert (tmp_path / "ctao-example.performance.fits").exists()
    assert (tmp_path / "ctao-example.event_selection.fits").exists()
